--[[
-- Copyright 2017 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]
-- [[
-- A module containing all the logic involved in transitioning through
-- combat that do not directly involve building the combat response.
-- ]]
local M = {}
local attack = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local pages = require 'pages'
local powers = require 'combat/powers'
local log = require 'log'
local scarlettables = require 'scarlettables'
local statuses = require 'combat/statuses'
local utilities = require 'utilities'

--[[
-- Given the name of the choice, returns a string representing the key
-- the user presses to select that option.
--
-- The key is selected by finding the character wrapped in parenthesis.
--
-- name - the name of the choice to extract an id from. The name must have
--  exactly one character (the key to press) wrapped in parenthesis
-- reservedchoiceids - A set of keys that are reserved for internal choices
-- (like the baskspace for undoing), defaults to the empty table
--
-- return a single character representing the keyboard key to press for the
-- specified choice
--]]
function M.extractchoiceid(name, reservedchoiceids)
    reservedchoiceids = reservedchoiceids or {}
    local openparens = assert(string.find(name, "%("), "No open parens in " .. name)
    local closeparens = assert(string.find(name, "%)"), "No close parents in " .. name)
    local choiceid = string.lower(
        string.sub(name, openparens + 1, closeparens - 1)
    )
    assert(
        not reservedchoiceids[choiceid],
        F"{choiceid} cannot be used for custom combat commands."
    )
    return choiceid
end

--[[
-- Given the current state of combat, returns the combat page to transition
-- to when the player has no more choices, and there is no page to transition
-- to automatically.
--
-- The page returned will be one of the following:
--  0. The choices for the current player, if that player still has options
--      available.
--  1. The choices for the next player
--  2. The page that decides the action for the next villain
--  3. The page that starts the next round
--
-- combatpage - The current state of combat
--
-- return The next combat page to transition to
-- ]]
function M.nextcombatpage(combatstate)
    local character = combatstateops.activecharacter(combatstate)
    local choices = combatstate.choicecounts[character.name]
    for _, choicename in ipairs(choices._keys) do
        local choice = choices[choicename]
        if choice.current > 0 then
            return character.choice
        end
    end
    return M.nextcharacterpage(combatstate)
end

-- [[
-- Given the current state of combat, returns the page for the next character.
-- ]]
function M.nextcharacterpage(combatstate)
    local charindex = combatstate.charindex + 1
    local activecharacter = combatstateops.activecharacter(combatstate)
    local char = combatstate.characters[charindex]
    while char and combatstateops.isdefeated(combatstate, char.name) do
        charindex = charindex + 1
        char = combatstate.characters[charindex]
    end
    if char then
        log.info("Next character: " .. char.name)
        log.info("Charindex: " .. charindex)
    else
        log.info("No character at " .. charindex .. " Probably moving on to next round.")
    end

    -- The round is over when we've made choices for every character in
    -- combat.
    if charindex > #combatstate.characters then
        log.info("All choices made. Resetting character index.")
        log.info("Moving on to the next round.")
        local combat = require 'combat/combat'
        local nextpage = combat.endround()
        local oldtransformer = nextpage.transformer
        function nextpage.transformer(combatstate1)
            combatstate1.charindex = 1
            statuses.endturn(activecharacter, combatstate1)
            oldtransformer(combatstate1)
        end
        return nextpage
    end
    local charactername = combatstate.characters[charindex].name
    local nextpage = scarlettables.copy(assert(
        combatstate.characters[charindex].choice,
        "Missing choice for " .. charactername
    ))
    local oldtransformer = nextpage.transformer
    function nextpage.transformer(combatstate1)
        statuses.endturn(activecharacter, combatstate1)
        oldtransformer(combatstate1)
        log.info("Not moving on to the next round, moving on to the next character.")
        combatstate1.charindex = charindex
        local character = combatstateops.activecharacter(combatstate1)
        statuses.startturn(character, combatstate1)
    end
    return nextpage
end

--[[
-- An operation on combatstate that sets up the state for the next round:
--  1. Increment the round counter by 1.
--  2. Drop the previous combatstate (backing up is not allowed across rounds)
--  3. Clear the action counts for the next round
--  4. Remove all defeated summons.
--  4. Reset the charindex to 1, allowing us to iterate over the characters
--  from the very beginning
--
--  combatstate: The state to prepare for the next round
--]]
function M.nextround(combatstate)
    log.info("Executing next round.")
    combatstate.round = combatstate.round + 1
    for _, character in pairs(combatstate.characters) do
        M.resetchoicecount(character, combatstate)
    end
    combatstate.charindex = 1
    local character = combatstateops.activecharacter(combatstate)
    while combatstateops.isdefeated(combatstate, character.name) do
        log.info("Looking for undefeated character.")
        combatstate.charindex = combatstate.charindex + 1
        if combatstate.charindex > #combatstate.characters then
            log.info("All characters defeated.")
            combatstate.charindex = 1
        end
        character = combatstateops.activecharacter(combatstate)
    end
    log.info("Transitioning to next round. Clearing backup.")
    combatstate.previous = nil
end

--[[
-- Resets the character's choice counts.
--]]
function M.resetchoicecount(character, combatstate)
    local choicecounts = combatstate.choicecounts[character.name]
    for _, choicename in ipairs(choicecounts._keys) do
        local choice = choicecounts[choicename]
        choice.current = choice.max
    end
end
--[[
-- Given a table of characters, and the set of all defeated characters, checks
-- if every character in the table has been defeated
--
-- characters - A table containing the characters to check for defeat as values
-- defeatedcharacters - The list of all characters who have been defeated
--
-- returns true if every character in characters has been defeated, false
-- otherwise
-- ]]
function M.charactersdefeated(characters, combatstate)
    local allDefeated = true
    for _, character in pairs(characters) do
        allDefeated = allDefeated
            and combatstateops.isdefeated(combatstate, character.name)
    end
    return allDefeated
end

--[[
-- Given the current state of combat, returns true if combat is over
-- (i.e. all the allies or enemies have been defeated), and false otherwise.
--
-- combatstate - The current state of combat
--
-- returns true if combat is over
-- ]]
function M.combatover(combatstate)
    return M.charactersdefeated(combatstate.allies, combatstate)
        or M.charactersdefeated(combatstate.enemies, combatstate)
end

--[[
-- An operation on gamestate that takes all the changes applied to characters
-- during combat (i.e. damage done) and applies them to the gamestate.
--
-- gamestate - The current state of the game
-- ]]
function M.applycombatchanges(gamestate)
    local combatstate = gamestate.combatstate
    for _, combatcharacter in pairs(combatstate.characters) do
        local gamecharacter = gamestate.characters[combatcharacter.name]
        -- We generate unique names for copies of generic enemies. The gamestate
        -- won't know about those characters, and we don't want to copy over
        -- their energy anyway.
        if gamecharacter then
            gamecharacter.statistics.energy = combatcharacter.statistics.energy
        end
    end
    gamestate.combatstate = nil
end


local arrows = {
    [utilities.up]="up",
    [utilities.down]="down",
    [utilities.right]="right",
    [utilities.left]="left",
    [utilities.backspace]="delete"
}

M.alphabet = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
    'm', 'n', 'o', 'p', 'q', 'r', 's', 't',  'u', 'v', 'w', 'x', 'y', 'z'}

M.alphabetset = {}
for i, letter in ipairs(M.alphabet) do
    M.alphabetset[letter] = i
end

-- Builds the table of choices to display for a standard combat page.
--
-- Note that `#` is treated as a special character when labeling the keys
-- for a choice. `#` represents numbers, and will automatically be expanded into
-- a choice for each digit, and one for each function key. This choice is
-- identical to the choice that `#`
-- maps to, except that the transformer also appends to `combatstate.selections`
-- the pressed digit (as a string) before invoking the original page's
-- transformer.
--
-- Furthermore '@' is treated as a special character when labeling keys,
-- except rather than representing digits, it represents characters.
--
-- Note that in order to display a choice with `(#)` without displaying a
-- separate choice for each digit, we rely on the
-- fact that the `_keys` are used to display the choices to the player, but that
-- when a choice is made, it is looked up directly. We store the choice with
-- `(#)` in the `_keys_` list (so that it is displayed), but we store the
-- actual expanded choices in the `choices` table:
--  {
--      _keys={"(#) Select a character"},
--      ["1"]=choice,
--      ["2"]=choice,
--      ...
--  }
--
-- combatpage - The page of combat to display
-- reservedchoiceids - A table of choice ids that are used by the engine, and
--  so can't be used by the developer
-- storypagebuilder - A function that takes a combatpage and returns a
--  gamepage to display it
--
-- returns A choices function on gamestate that takes the gamestate and returns
--  an ordered table that maps strings to story pages showing combat
--]]
function M.buildchoices(combatpage, reservedchoiceids)
    return function(gamestate)
        local combatstate = gamestate.combatstate
        local combatchoices = {_keys={}}
        local choices = combatpage.choices(gamestate.combatstate)
        for _, name in ipairs(choices._keys) do
            local choiceid = M.extractchoiceid(name, reservedchoiceids)
            choiceid = arrows[choiceid] or choiceid
            log.trace("Extracted choiceid: " .. choiceid)
            if reservedchoiceids[choiceid] then
                local errorstring =
                    F"The choice id '{choiceid}' is reserved for engine use."
                log.error(errorstring)
                assert(false, errorstring)
            end
            if choiceid == '#' then
                if LOVE then
                    -- Enemy selection
                    for j=1,9 do
                        local key = tostring(j)
                        combatchoices[key] = M.registerselection(
                            key,
                            M.page(choices[name])
                        )
                        table.insert(combatchoices._keys, key)
                    end
                    -- Ally selection
                    for j=1,12 do
                        local key = "f" .. tostring(j)
                        combatchoices[key] = M.registerselection(
                            key,
                            M.page(choices[name])
                        )
                        table.insert(combatchoices._keys, key)
                    end
                else
                    for j=1,#combatstate.characters do
                        local key = tostring(j)
                        combatchoices[key] = M.registerselection(
                            key,
                            M.page(choices[name])
                        )
                        table.insert(combatchoices._keys, key)
                    end
                end
            elseif choiceid == '@' then
                for _,char in ipairs(M.alphabet) do
                    if not reservedchoiceids[char] then
                        combatchoices[char] = M.registerselection(
                            char,
                            M.page(choices[name])
                        )
                        table.insert(
                            combatchoices._keys,
                            char
                        )
                    end
                end
            else
                combatchoices[name] = M.page(choices[name])
                table.insert(combatchoices._keys, name)
            end
        end
        M.addmetadatachoices(combatchoices, gamestate, reservedchoiceids)
        return combatchoices
    end
end

--[[
-- Modifies choices 1-9 and F1-F12 to display character information in the
-- metalog when selected. A character's data is displayed in the following
-- format:
--  name: strength: strength speed: speed power: power
--  powers: power,names
--]]
function M.addmetadatachoices(combatchoices, gamestate, reservedchoiceids)
    local combatstate = gamestate.combatstate
    if LOVE then
        for i=0,9 do
            if combatstate.enemies[i] then
                local key = tostring(i)
                local combatpage = combatchoices[key]
                if not combatpage then
                    local choice = pages.page()
                    function choice.body(gamestate)
                        local enemy = combatstate.enemies[i]
                        return {
                            "Viewing " .. enemy.combatname,
                            "Select a power to view its description."
                        }
                    end
                    function choice.choices(gamestate)
                        local combatstate = gamestate.combatstate
                        local enemy = combatstate.enemies[i]
                        if enemy then
                            return M.charactermetadatachoices(
                                enemy,
                                gamestate,
                                choice,
                                reservedchoiceids
                            )
                        end
                        return {_keys={}}
                    end
                    function choice.transformer(gamestate)
                        local combatstate = gamestate.combatstate
                        local enemy = combatstate.enemies[i]
                        if enemy then
                            combatstate.metalog = M.charactermetadata(enemy, reservedchoiceids)
                        end
                    end
                    combatchoices[key] = choice
                    table.insert(combatchoices._keys, key)
                end
            end
        end
        for i=1,12 do
            if combatstate.allies[i] then
                local key = "f" .. i
                local combatpage = combatchoices[key]
                if not combatpage then
                    local choice = pages.page()
                    function choice.body(gamestate)
                        local ally = combatstate.allies[i]
                        return {
                            "Viewing " .. ally.combatname,
                            "Select a power to view its description."
                        }
                    end
                    function choice.choices(gamestate)
                        local combatstate = gamestate.combatstate
                        local ally = combatstate.allies[i]
                        if ally then
                            return M.charactermetadatachoices(
                                ally,
                                gamestate,
                                choice,
                                reservedchoiceids
                            )
                        end
                        return {_keys={}}
                    end
                    function choice.transformer(gamestate)
                        local combatstate = gamestate.combatstate
                        local ally = combatstate.allies[i]
                        if ally then
                            combatstate.metalog = M.charactermetadata(ally, reservedchoiceids)
                        end
                    end
                    combatchoices[key] = choice
                    table.insert(combatchoices._keys, key)
                end
            end
        end
    else
        for i=1,#combatstate.characters do
            local key = tostring(i)
            local combatpage = combatchoices[key]
            if not combatpage then
                local choice = pages.page()
                function choice.body(gamestate)
                    local character = combatstate.characters[i]
                    return {
                        "Select a power to view its description below."
                    }
                end
                function choice.choices(gamestate)
                    local combatstate = gamestate.combatstate
                    local character = combatstate.characters[i]
                    if character then
                        return M.charactermetadatachoices(
                            character,
                            gamestate,
                            choice,
                            reservedchoiceids
                        )
                    end
                    return {_keys={}}
                end
                function choice.transformer(gamestate)
                    local combatstate = gamestate.combatstate
                    local character = combatstate.characters[i]
                    if character then
                        combatstate.metalog = M.charactermetadata(character, reservedchoiceids)
                    end
                end
                combatchoices[key] = choice
                table.insert(combatchoices._keys, key)
            end
        end
    end
end

-- [[
-- Returns a table of game choices that allow the player to look at information
-- about a character's powers.
--
-- character: The character being examined
--
-- return The list of choices available in this screen that allow the player
-- to read up on that character's powers
-- ]]
function M.charactermetadatachoices(character, gamestate, thispage, reservedchoiceids)
    local powers = character.powers
    local choices = {_keys={}}
    M.addmetadatachoices(choices, gamestate, reservedchoiceids)
    local letterkey = 0
    if powers._keys then
        local scarlettables = require 'scarlettables'
        log.info("Powers: " .. scarlettables.string(powers))
        for i, powername in ipairs(powers._keys) do
            letterkey = letterkey + 1
            while reservedchoiceids[M.alphabet[letterkey]] do
                letterkey = letterkey + 1
            end
            local key = M.alphabet[letterkey]
            log.info(F"Getting description for {powername}.")
            log.info(F"Power: {powers[powername]}.")
            choices[key] = M.getpowerdescription(
                powername,
                powers[powername],
                character,
                key,
                thispage,
                reservedchoiceids
            )
            table.insert(choices._keys, key)
        end
    end
    return choices
end


local function canusepowers(combatstate, character)
    local choicecounts = combatstate.choicecounts
    local powercount = choicecounts[character.name].power
    local isactive = combatstateops.activecharacter(combatstate).name
        == character.name
    return isactive and powercount.current > 0
end

-- [[
-- Returns a game page that displays a description of the power, and allows
-- the player to use it, or go back.
--
-- The power is displayed in the following format:
-- name:
--  description
--  cost: power.cost
--  Num targets: power.numtargets
--  Statistics: power.stats
--
--
-- powername - The name of the power to display
-- power - The power to display
-- character - The character whose power is being examined
-- key - The KeyConstant representing the keyboard key pressed to view this power
-- previouspage - The page to go back to
--
-- return A page that prints a list of strings that contain the above description of the power
-- ]]
function M.getpowerdescription(powername, power, character, key, previouspage, reservedchoiceids)
    local page = pages.page()
    function page.body(gamestate)
        local combatstate = gamestate.combatstate
        log.info(F"Displaying description for '{powername}'.")
        local numtargets = power.numtargets
        numtargets = numtargets == powers.ALL_TARGETS
            and "all"
            or numtargets == powers.ALL_OPPONENTS
            and "all opponent"
            or numtargets == powers.ALL_ALLIES
            and "all ally"
            or tostring(numtargets)
        local powerdescription = power.description
        if type(powerdescription) == "function" then
            powerdescription = powerdescription(character)
        end
        local statistics = #power.stats == 1 and "statistic" or "statistics"
        local cost = type(power.cost) == "number"
            and power.cost
            or power.cost(character)
        local description = {
            "*" .. powername .. "*" .. ": ",
            powerdescription,
            "  Costs *" .. cost .. "* energy.",
            "  Targets *" .. numtargets .. "* targets.",
            "  Uses *"
                .. table.concat(power.stats, ", ")
                .. "* "
                .. statistics
                .. "."
        }
        if canusepowers(combatstate, character) then
            table.insert(description, "(Enter) Use " .. powername)
        end
        return {table.concat(description, " ")}
    end
    function page.choices(gamestate)
        local choices = M.charactermetadatachoices(
            character,
            gamestate,
            previouspage,
            reservedchoiceids
        )
        if canusepowers(gamestate.combatstate, character) then
            local powercopy = scarlettables.copy(power)
            local oldtransformer = powercopy.transformer
            function powercopy.transformer(combatstate)
                oldtransformer(combatstate)
                local choicecounts = combatstate.choicecounts[character.name]
                local powercount = choicecounts.power
                powercount.current = powercount.current - 1
            end
            choices["return"] = M.page(powercopy)
            table.insert(choices._keys, "return")
        end
        return choices
    end
    return page
end

local function statrepresentation(basevalue, modifiers)
    local representation = {tostring(basevalue)}
    for name, modifier in pairs(modifiers) do
        if name ~= "damage" then
            local modstring = modifier >= 0 
                and F"+{modifier}"
                or tostring(modifier)
            table.insert(representation, F"{modstring}({name})")
        end
    end
    return representation
end

--[[
-- Given a character, returns a brief description of the character.
--
-- character: The character to get a brief description of
--
-- return A list of strings describing the specified character
--]]
function M.charactermetadata(character, reservedchoiceids)
    local powernames = {}
    if character.powers._keys then
        local letterkey = 0
        for i, powername in ipairs(character.powers._keys) do
            letterkey = letterkey + 1
            while reservedchoiceids[M.alphabet[letterkey]] do
                letterkey = letterkey + 1
            end
            local key = M.alphabet[letterkey]
            powernames[i] = "(" .. key .. ")" .. powername
        end
    end
    local statusnames = {}
    for _, status in ipairs(character.statuses) do
        table.insert(statusnames, status.status.name)
    end
    return
        {
            
            "**** Info: " .. character.combatname .. " ****",
            table.concat({
                "  strength: " .. attack.statistic(character, "strength"),
                "  speed: " .. attack.statistic(character, "speed"),
                "  power: " .. attack.statistic(character, "power")
                },
            " "),
            table.concat({
                "  attack: " .. attack.attackstat(character),
                "  defense: " .. attack.defensestat(character),
                "  shield: " .. attack.shield(character),
                "  retaliation: " .. attack.retaliation(nil, character)
                },
            " "),
            "statuses: " .. table.concat(statusnames, ", "),
            "powers: " .. table.concat(powernames, ", ")
        }
end

--[[
-- Given a Love KeyConstant, and the page to transition to if that selection
-- is made, returns a new page that registers the selection with
-- `gamestate.combatstate.selections` and then automatically transitions to the
-- selected page.
--
-- selection: The Love KeyConstant that needs to be registered
-- selectedpage: The page to transition to if the selection is made
--
-- return: A page that registers the selection with `combatstate.selections`
--  and then transitions to the specified page
-- ]]
function M.registerselection(selection, selectedpage)
    return pages.page(
        nil,
        function()
            return selectedpage, selection
        end,
        nil,
        function(gamestate)
            local selections = gamestate.combatstate.selections
            for _, oldselection in pairs(selections) do
                if oldselection == selection then
                    return
                end
            end
            table.insert(selections, selection)
        end
    )
end

--[[
-- Given the combat page to display, returns a function on gamestate that
-- returns the next gamepage to transition to, if any.
--
-- combatpage - The combat page that needs to be displayed
-- gamepagebuilder - A function that takes a combatpage and returns the
-- associated gamepage
-- endcombat - A game page that ends combat
--
-- return A function on combatpage that returns the next game page to transition
-- to automatically, if any
-- ]]
function M.buildautomatics(combatpage, endcombat)
    return function(gamestate)
        local combatstate = gamestate.combatstate
        if M.combatover(combatstate) then
            return endcombat
        end
        local nextcombatpage = combatpage.automatics(combatstate)
        local choices = combatpage.choices(combatstate)
        local firstchoice
        if choices._keys then
            firstchoice = next(choices._keys)
        end
        if not nextcombatpage and firstchoice == nil then
            nextcombatpage = M.nextcombatpage(combatstate)
        end
        return nextcombatpage and M.page(nextcombatpage)
    end
end

--[[
-- Given a combat page, builds the transformer for the associated game page.
-- In addition to applying the combatpage's transformer, this transformer will
-- also automatically increment the combat state's char
-- index (i.e. which character we are selecting an action for) if the
-- current combat page does not have any child pages.
--
-- The transformer also appends the body of the combat page to the
-- combatstate's log.
--
-- combatpage - The combat page to display
--
-- return An operation on gamestate that applies the changes in
--  combatpage.transformer and transitions to the next character when necessary
-- ]]
function M.buildtransformer(combatpage)
    return function(gamestate)
        local combatstate = gamestate.combatstate
        combatpage.transformer(combatstate, combatpage)
        local choices = combatpage.choices(combatstate) or {_keys={}}
        local combatbody = combatpage.body(combatstate)
        local combatlog = combatstate.combatlog
        local loglength = #combatlog

        for i, paragraph in ipairs(combatbody) do
            combatlog[loglength + i] = paragraph
        end
        local character = combatstate.characters[combatstate.charindex]
        local metalog = combatstate.metalog
        if choices._keys and character.pc then
            for i, choice in ipairs(choices._keys) do
                if i % 2 == 0 then
                    metalog[#metalog] = metalog[#metalog]
                        .. " "
                        .. choice
                else
                    table.insert(metalog, choice)
                end
            end
        end
        combatstate.currentpage = combatpage
    end
end

--[[
-- Increments to the index of the next character who has not been defeated.
--
-- combatstate - The current state of the combat
-- charindex - The current char index
--
-- return The index of the next character who has not been defeated, or the 
-- first character if everyone has been defeated
-- ]]
function M.incrementcharindex(combatstate, charindex)
    local nextindex = charindex
    local numcharacters = #combatstate.characters
    repeat
        local character = combatstate.characters[nextindex]
        local isdefeated = combatstateops.isdefeated(
            combatstate,
            character.name
        )
        if isdefeated then
            nextindex = nextindex + 1
        end
    until not isdefeated or nextindex > numcharacters
    return nextindex > numcharacters and 1 or nextindex
end

--[[
--A function that builds the choices function for the end of combat.
--
-- restartkey -- The key to use to restart combat
-- continueky -- The key to use to continue the game
-- storypagebuilder - A function that takes the name of a story page, and
-- a transformer and returns a new story page
--
--returns A function on gamestate that returns the table of choices to display
--  at the end of combat
--]]
function M.endcombatchoices(restartkey, continuekey)
    return function(gamestate)
        local combatstate = gamestate.combatstate
        local choices = {_keys={}}
        if M.charactersdefeated(combatstate.allies, combatstate) then
            if combatstate.defeat then
                choices[continuekey] = pages.page(
                    nil,
                    function() return combatstate.defeat end,
                    nil,
                    M.applycombatchanges
                )
                table.insert(choices._keys, continuekey)
            end
        elseif M.charactersdefeated(combatstate.enemies, combatstate) then
            choices[continuekey] = pages.page(
                nil,
                function() return combatstate.victory end,
                nil,
                M.applycombatchanges
            )
            table.insert(choices._keys, continuekey)
        end

        local startcombatpage = M.startcombatpage(
            gamestate,
            combatstate.victory,
            combatstate.defeat,
            combatstate.start.allies,
            combatstate.start.enemies
        )

        choices[restartkey] = startcombatpage
        table.insert(choices._keys, restartkey)
        return choices
    end
end

--[[
-- Builds a game page that starts combat.
--
-- gamestate: The current state of the game
-- victory: The story event to transition to if the player wins
-- loss: The optional story event to transition to if the player loses
-- allies: The characters controlled by the player
-- enemies: The characters the player is fighting
--
-- return A gamepage that starts combat with a fresh combat state
--]]
function M.startcombatpage(gamestate, victory, loss, allies, enemies)
    local combatstate = combatstateops.combatstate(
        gamestate,
        victory,
        loss,
        allies,
        enemies
    )
    if loss then
        table.insert(
            combatstate.combatlog,
            "This battle is *optional* meaning you don't have to win to continue the story."
        )
    else
        table.insert(
            combatstate.combatlog,
            "Things are looking pretty grim! Gotta win this one to continue the story."
        )
    end
    table.insert(
        combatstate.combatlog,
        "Select a character (including enemies) to view their stats and powers!"
    )
    return pages.page(
        nil,
        function() return M.page(combatstate.choice) end,
        nil,
        function(gamestate1) gamestate1.combatstate = combatstate end
    )
end

--[[
-- Builds the game page at the end of combat
-- that offers the player the ability to either continue the game (if possible)
-- or restart combat.
--]]
local function endcombat()
    local function transformer(gamestate)
        local combatstate = gamestate.combatstate
        combatstate.metalog = {}
        if combatstateops.alldefeated(combatstate, combatstate.allies) then
            table.insert(
                combatstate.combatlog,
                "Oh no, you've been defeated! Who will save us now?"
            )

            if combatstate.defeat then
                table.insert(combatstate.metalog, "(C)ontinue")
            end
        elseif combatstateops.alldefeated(combatstate, combatstate.enemies) then
            table.insert(
                combatstate.combatlog,
                "More villains brought to justice with grace and aplomb!"
            )
            table.insert(combatstate.metalog, "(C)ontinue")
        else
            assert(false, "Somebody should have been defeated!")
        end
    end
    local choices = M.endcombatchoices("r", "c")

    return pages.page(
        function(gamestate) return gamestate.combatstate.combatlog end,
        nil,
        choices,
        transformer
    )
end

local END_COMBAT = endcombat()

--[[
-- Given a combat page, returns a gamepage that displays the current state
-- of combat and presents the passed in choice to the player.
--
-- This function assumes that all the combat pages are keyed by strings of
-- the form (C)hoice, or c(h)oice. In other words, the key that is used to
-- select the choice is wrapped in parenthesis.
--
-- combatpage - The choice that needs to be presented to the player
--
-- returns A game page that displays the specified combatpage
--]]
function M.page(combatpage)
    local reservedchoiceids = {
        e=1,
        l=1,
        t=1,
        backspace=1
    }
    local combatchoices = M.buildchoices(combatpage, reservedchoiceids)
    local newpage = pages.page(
        M.body,
        M.buildautomatics(combatpage, END_COMBAT),
        combatchoices,
        M.buildtransformer(combatpage)
    )
    if combatpage.name then
        newpage.combatname = "game;" .. combatpage.name
    end
    return newpage
end

--[[
-- Given the current state of the game, builds a list of paragraphs that
-- describe the current state of combat.
--
-- The body consists of: the log of the entire combat to this point, plus
-- a compact overview of the enemies and allies currently involved in combat.
--
-- gamestate: The current state of the game
-- combatpage: The current page of combat
--
-- return: A list of paragraphs describing the current state of combat
-- ]]
function M.body(gamestate)
    return gamestate.combatstate.combatlog
end



--[[
-- Given a list of paragraphs containing the current combat combatlog, and a list
-- of paragraphs containing the text to display for the current choice,
-- determines if the combatlog already ends with the choice body
-- ]]
function M.alreadydisplayed(combatlog, body)
    local bodylength = #body
    local combatloglength = #combatlog
    for i = 0, bodylength - 1 do
        if combatlog[combatloglength-i] ~= body[bodylength-i] then
            return false
        end
    end
    return true
end


--[[
-- Given a list of characters, returns a table that maps character name to the
-- string that represents that character in combat.
--
-- characters - The characters to build labels for
-- keyprefix - A string to put in front of the character's shortcut key label
--  (i.e. "F" if you want to use the function keys), defaults to ""
-- concise - A boolean, true if the concise combat board should be displayed,
-- false if the full board should be displayed
--
-- returns A table from character name to the string representing that
--  character in combat
-- ]]
function M.buildcharacterlabels(characters, keyprefix, concise)
    keyprefix = keyprefix or ""
    local labels = {}
    -- (key)name:statuses:energy/maxenergy
    local labeltemplate = "(%s)%s:%d/%d:%s"
    for i, character in ipairs(characters) do
        labels[character.name] = string.format(
            labeltemplate,
            keyprefix .. i,
            concise and character.concisename or character.combatname,
            math.floor(character.statistics.energy),
            character.statistics.maxenergy,
            M.buildstatuslabels(character.statuses, concise)
        )
    end
    return labels
end

--[[
-- Given a list of statuses, returns a string representing the statuses for
-- display during combat
--
-- statuses - The list of statuses inflicting a character. Each status is
--  a table with the following keys:
--      status - A table that defines the status currently afflicting a
--          character. This table must include the a key `display` that
--          contains a string used to display the status in combat.
--      duration - A number representing how many rounds the character is
--          inflicted
-- concise - A boolean, true if the concise board should be displayed, false
--  otherwise
--
--  returns A string of the format s1(d1),s2(d2),...sn(dn) where s1...sn are
--  characters representing each status, and d1...dn are the durations of each
--  status
-- ]]
function M.buildstatuslabels(statuses, concise)
    local statuslabel = {}
    for i, status in ipairs(statuses) do
        statuslabel[i] = string.format(
            "%s(%d)",
            concise and status.status.concisename or status.status.display,
            status.duration
        )
    end
    return table.concat(statuslabel, ",")
end

return M
