--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
-- [[
-- This module contains the definitions of the custom choices available in
-- Scarlet Moon.
-- ]]
local M = {}

local attack = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powerutils = require 'data/powerutils'
local statusops = require 'combat/statuses'
local targeting = require 'combat/target'
local utilities = require 'combat/combatutils'

--[[
-- Given the attacker, defender and the current state of combat, returns
-- the damage that the user will do against the target, and a single paragraph
-- describing the attack.
--
-- This is very similar to the same-named function in combat/attack, except
-- its base damage is based on the user's power, rather than their strength,
-- and it doesn't trigger retaliation (though it can be blocked by shields).
--
-- user: The shooter
-- target: The target of the shot
-- combatstate: The current state of combat
--
-- return The damage done, 0 for retaliation, and a list of strings describing 
-- the action
--]]
local function shootdamage(user, target, combatstate)
    local basedamage = attack.attackstat(user)
    local defense = attack.defensestat(target)
    local damagedone = attack.shield(target, math.max(1, basedamage - defense))
    local text = {
        F"{user.combatname} shoots {target.combatname} for {damagedone} damage!"
    }
    return damagedone, nil, text
end

local function HeShe(character)
    return character.attributes.gender == "female" and "She" or "He"
end

local function hisher(character)
    return character.attributes.gender == "female" and "her" or "his"
end

--[[
-- Shoot is a lot like attack, except that it uses power rather than strength
-- as its base damage. Relative speed is still used for defense.
--]]
M.shoot = targeting.target(
    function(combatstate, previouspage, targetkey)
        return attack.attack(combatstate, previouspage, shootdamage, targetkey)
    end,
    1
)


M.attack = targeting.target(
    function(combatstate, previouspage, targetkey)
        return attack.attack(combatstate, previouspage, nil, targetkey)
    end,
    1
)

local function reducebuff(buff, target)
    return function(primitive, combatstate)
        local spankerstrength = combatstateops.findcharacter(
            combatstate,
            primitive.username
        ).statistics.strength
        local target = combatstateops.findcharacter(
            combatstate,
            primitive.targetname
        )
        local spankeepower = target.statistics.power
        local duration = math.max(1, spankerstrength - spankeepower)
        local description = {
            string.format(
                "%s's spanking has reduced %s by %d turns!",
                target.combatname,
                buff.name,
                duration
            )
        }
        return buff, -duration, description
    end
end

function M.spankreducebuff(buff, target)
    return reducebuff(buff, target)
end

local function spankingtarget(combatstate, targetkey)
    assert(targetkey, "Targetkey must be non-nil for a spanking.")
    local target = utilities.resolvetargetkey(combatstate, targetkey)
    if not target then
        return _, pages.page(
            function()
                return {"Unknown target: " .. targetkey}
            end,
            function()
                return previouspage
            end
        )
    end
    return target
end

-- [[
-- Spanking affects the abilities of some enemies. Furthermore,
-- when a spanking is applied, it reduces the duration of the target's best
-- buff, and cuts their retaliation in half.
--
-- descriptionsfunction - A function that takes the user and target and returns a list of
-- strings to randomly choose from when administering a spanking. If not provided, defaults
-- to a list of hard-coded strings that assumes the spanker is a human-sized person.
-- ]]
function M.spank(descriptionsfunction) 
    return targeting.target(
        function(combatstate, previouspage, targetkey)
            local spanker = combatstateops.activecharacter(combatstate)
            local target, errorpage = spankingtarget(combatstate, targetkey)
            if not target then
                return errorpage
            end
            local possibledescriptions
            if descriptionsfunction then
                possibledescriptions = descriptionsfunction(spanker, target)
            else
                possibledescriptions = {}
                if target.huge then
                   table.insert(
                       possibledescriptions,
                       string.format(
                           "%s whips past %s and smacks %s hard on the ass. %s yelps"
                           .. " and clutches %s bottom.",
                           spanker.combatname,
                           target.combatname,
                           powerutils.himher(target),
                           target.combatname,
                           powerutils.hisher(target)
                       )
                   )
                else
                    if powerutils.issummon(target) then
                        table.insert(
                            possibledescriptions,
                            F(
                                "{spanker.combatname} smashes {target.combatname} "
                                .. "on the closest thing {target.combatname} has to "
                                .. "a bottom. {target.combatname} wavers unsteadily "
                                .. " as the spanking disrupts it!"
                            )
                        )
                    else
                        table.insert(
                            possibledescriptions,
                            F(
                                "{spanker.combatname} bends {target.combatname} over "
                                .. "{powerutils.hisher(spanker)} knee. "
                                .. "{target.combatname} "
                                .. "squawks indignantly and kicks {powerutils.hisher(target)} "
                                .. "feet. {spanker.combatname} peppers "
                                .. "{target.combatname}'s round bottom, "
                                .. "making the "
                                .. "vulnerable cheeks dance. {target.combatname} kicks "
                                .. "and squeals."
                            )
                        )
                        table.insert(
                            possibledescriptions,
                            F(
                                "{spanker.combatname} tucks {target.combatname} under "
                                .. "{powerutils.hisher(spanker)} arm and whales on "
                                .. "{target.combatname}'s bucking bottom, sending ripples "
                                .. "across the sensitive cheeks. {target.combatname} yelps "
                                .. "and squirms."
                            )
                        )
                    end
                end
            end
            return pages.page(
                nil, 
                nil,
                nil, 
                function() 
                    local index = math.random(#possibledescriptions)
                    local description = possibledescriptions[index]
                    table.insert(combatstate.combatlog, description)
                    local status = require("data/statuses").spanked
                    statusops.addstatus(target, status, 3, spanker, nil, combatstate)
                end
            )
        end,
        1
    )
end

return M
