--[[
-- Copyright 2018 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
-- ]]
--
-- [[
-- A list of all the statuses in the game. See combat/combat.md#Status for
-- details on statuses.
-- Combatstate has a special mapping `statusstate` that maps a status name
-- to whatever additional information the status needs in order to reverse
-- itself when the status expires. For example, a
-- status that increases a character's strength may have
-- `combatstate.statusstate = { character.name=bonus}
-- which tracks the bonus applied to that character.
-- ]]
local M = {}

local attacks = require 'combat/attack'
local combatstateops = require 'combat/combatstate'
local combatutils = require 'combat/combatutils'
local F = require 'f-strings/F'
local log = require 'log'
local pages = require 'pages'
local powerutils = require 'data/powerutils'
local scarletchoices = require 'data/scarletchoices'
local scarletfunctions = require 'scarletfunctions'
local scarletmoonstatuses = require 'data/scarletmoonstatuses'
local scarlettables = require 'scarlettables'
local statusops = require 'combat/statuses'

for name, status in pairs(scarletmoonstatuses) do
    M[name] = status
end

-- [[
-- Removes the specified choice from the provided ordered table of choices.
--
-- choices: The ordered table of choices to remove a choice from
-- choicename: The name of the choice to remove
-- ]]
local function removechoice(choices, choicename)
    choices[choicename] = nil
    scarlettables.remove(choices._keys, choicename)
end


--[[
-- Given a list of opponents, returns the opponent with the highest energy.
--]]
local function highestenergytarget(opponents)
    local _, target = next(opponents)
    for _, opponent in ipairs(opponents) do
        if opponent.statistics.energy > target.statistics.energy then
            target = opponent
        end
    end
    return target
end

--[[
-- Provides a function that abstracts away the most common pattern for applying statuses:
--
-- 1. Get the current version of this status that's afflicting the target.
-- 2. If there is such a status, stack it.
-- 3. If not, and the duration is positive, apply the status.
-- 4. Otherwise, if after stacking the status' duration is 0 or les, remove
-- the status
--
-- target: The character to apply the status to
-- status: The status to apply
-- stack: A transformer on statusduration, and stacks the status
-- apply: A transformer on character that applies the effects of the
--  status to the character
-- remove: A transformer on character that reverses the effects of the
--  status
--]]
local function applystatus(target, status, stack, apply, remove, duration)
    local statusduration = statusops.statusduration(target, status)
    if statusduration then
        stack(statusduration)
    end
    if not statusduration and duration > 0 then
        statusops.addstatus(target, status, duration)
        apply(target)
    elseif statusduration and statusduration.duration <= 0 then
        statusops.removestatus(target, status)
        remove(target, statusduration)
    end
end

--[[
-- A function that returns a stacking function meant to be used
-- with applystatus that increases the duration of the status.
--
-- duration: The amount to increase the duration of a status by
--
-- returns A transformer on statusduration that increases the
-- duration by `duration`.
--]]
local function increaseduration(duration)
    return function(statusduration)
        statusduration.duration = statusduration.duration + duration
    end
end

------------------- Away -------------------------------------
--[[
-- The away status reduces the target's maximum (and current) attack count by 1
-- for the duration of the status. Note that the attack count can still be
-- modified one way or another by other statuses.
--
--]]
M.away = {
    name="Away",
    display="away",
    concisename="A",
    buff=-1,
    apply=function(target, combatstate)
        local attackcount = combatstate.choicecounts[target.name].attack
        if attackcount then
            attackcount.max = attackcount.max - 1
            attackcount.current = attackcount.current - 1
        end
    end,
    remove=function(target, combatstate)
        local attackcount = combatstate.choicecounts[target.name].attack
        if attackcount then
            attackcount.max = attackcount.max + 1
            attackcount.current = attackcount.current + 1
        end
    end
}

------------------- Chainwhip -------------------------------------
-- [[
-- Chainwhip increases the target's attack by the value of their speed
-- statistic.
-- ]]

M.chainwhip = {
    name="Chainwhip",
    display="chainwhip",
    concisename="CW",
    buff=2
}

function M.chainwhip.apply(character, combatstate)
    local buff = attacks.statistic(character, "speed")
    character.secondarystatistics.attackmodifiers[M.chainwhip.name] = buff
end

function M.chainwhip.remove(character, combatstate)
    character.secondarystatistics.attackmodifiers[M.chainwhip.name] = nil
end

---------------------- Leap -------------------------
--[[
-- Grants a character the ability to jump far distances, allowing them
-- to recover from Away at the start of their turn.
--]]

M.leap = {
    name="Leap",
    display="leap",
    concisename="L",
    buff=1,
    apply=function() end,
    remove=function() end
}

function M.leap.startturn(character, combatstate)
    if statusops.statusduration(character, M.away) then
        statusops.removestatus(character, M.away, combatstate)
        table.insert(
            combatstate.combatlog,
            "Buzzsaw leaps back into the fray!"
        )
    end
end

----------------- Drain ----------------------------------
--[[
-- Drain reduces the target's power by 1. The key difference between drain and
-- most other statuses is that Drain's stat reduction stacks. Rather than just
-- increasing the duration of the status, the status also reduces power by 1
-- with each new application.
--]]

M.drain = {
    name="Drain",
    display="drain",
    concisename="D",
    buff=-1
}

function M.drain.apply(character, combatstate)
    local powermodifiers = character.secondarystatistics.powermodifiers
    powermodifiers[M.drain.name] = -1
end

function M.drain.stack(character, combatstate)
    local powermodifiers = character.secondarystatistics.powermodifiers
    local penalty = powermodifiers[M.drain.name]
    powermodifiers[M.drain.name] = penalty - 1
end

function M.drain.remove(character, combatstate)
    character.secondarystatistics.powermodifiers[M.drain.name] = nil
end

-------------------------- OTK ----------------------------------------
--[[
-- OTK drops all of a character's actions to zero.
--]]

M.otk = {
    name="OTK",
    display="otk",
    concisename="OTK",
    buff=-10
}

function M.otk.apply(target, combatstate)
    if not combatstate.statusstate[M.otk] then
        combatstate.statusstate[M.otk] = {}
    end
    combatstate.statusstate[M.otk][target.name] = {}
    local statusstate = combatstate.statusstate[M.otk][target.name]
    local choicecounts = combatstate.choicecounts[target.name]
    for _, actionname in ipairs(choicecounts._keys) do
        statusstate[actionname] = choicecounts[actionname].max
        choicecounts[actionname].max = 0
        choicecounts[actionname].current = 0
    end
end

function M.otk.remove(target, combatstate)
    local statusstate = combatstate.statusstate[M.otk][target.name]
    combatstate.statusstate[M.otk][target.name] = nil
    local choicecounts = combatstate.choicecounts[target.name]
    for name, max in pairs(statusstate) do
        local choice = choicecounts[name]
        if choice then
            choice.max = choice.max + max
            choice.current = choice.max
        end
    end
end

function M.otk.startturn(character, combatstate)
    table.insert(
        combatstate.combatlog, 
        character.combatname .. " squirms and tries to escape!"
    )
end

--------------------------- Spanked -------------------------------------
--[[
-- Spanked reduces the duration of the target's best buff upon being applied,
-- possibily even removing it completely. Its presence can also interfere
-- with some characters' powers.
-- Spanked reduces the target's best buff by a duration equal to the user's
-- power.
--]]

M.spanked = {
    name="Spanked",
    display="spanked",
    concisename="S",
    buff=-1
}

function M.spanked.apply(target, combatstate, user)
    local bestbuff = statusops.bestbuff(target)
    if bestbuff then
        local power = attacks.statistic(user, "power")
        local spankdefense = math.floor(attacks.statistic(target, "power") / 2)
        local penalty = math.max(2, power - spankdefense)
        statusops.decrementstatus(bestbuff, target, penalty, combatstate)
        table.insert(
            combatstate.combatlog,
            "The spanking reduces the  duration of " 
                .. target.combatname .. "'s "
                .. bestbuff.name .. " by " .. penalty .. " turns!"
        )
        local retaliationmodifiers = target.secondarystatistics.retaliation
        target.secondarystatistics.retaliation.spanked = -power
        table.insert(
            combatstate.combatlog,
            "The spanking reduces  " 
                .. target.combatname .. "'s "
                .. "shield and retaliation (if any) by " .. power .. "!"
        )
        target.secondarystatistics.shield.spanked = -power
    end
end

function M.spanked.remove(target)
    target.secondarystatistics.retaliation.spanked = nil
    target.secondarystatistics.shield.spanked = nil
end

--------------------------- Succuspanked -------------------------------------
--[[
-- A variation of Spanked whose buff disruption is based on power rather
-- than strength
--]]

M.succuspanked = {
    name="Succuspanked",
    display="succuspanked",
    concisename="SS",
    buff=-1
}

function M.succuspanked.apply(character, combatstate, user)
    local bestbuff = statusops.bestbuff(character)
    if bestbuff then
        local penalty = math.floor(attacks.statistic(user, "power") / 2)
        table.insert(
            combatstate.combatlog,
            "The spanking reduces the  duration of " 
                .. character.combatname .. "'s "
                .. bestbuff.name .. " by " .. penalty .. " turns!"
        )
        statusops.decrementstatus(bestbuff, character, penalty, combatstate)
    end
end

--------------------- Frozen -----------------------------------------
--[[
-- For anybody but the supervillain Poseidon, being Frozen reduces a character's
-- attack and defense statistics to 0.
--
-- Poseidon on the other hand has his attack and defense increased by the 
-- duration / 2, and has damage reduction of the same amount.
--]]

M.frozen = {
    name="Frozen",
    display="frozen",
    concisename="Fr",
    buff=function(character)
        local manipulateswater = character.combatname == "Poseidon" 
            or character.combatname == "Hydrogator"
        return manipulateswater and 9 or -20
    end
}

function M.frozen.apply(target, combatstate, user, potency)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    if M.frozen.buff(target) > 0 then
        attackmodifiers[M.frozen.name] = potency
        defensemodifiers[M.frozen.name] = potency
    else
        attackmodifiers[M.frozen.name] = -attacks.attackstat(target)
        defensemodifiers[M.frozen.name] = -attacks.defensestat(target)
    end
end

function M.frozen.remove(target, combatstate)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    attackmodifiers[M.frozen.name] = nil
    defensemodifiers[M.frozen.name] = nil
end

--------------------- Wet -----------------------------------------
--[[
-- Wet doesn't do anything directly. Instead, Poseidon needs to be inflicted
-- with it for many of his powers to trigger, and his water-based powers do
-- extra damage to anyone inflicted with it.
--
-- Similarly, it also cancels out warm and aflame.
--]]

M.wet = {
    name="wet",
    display="Wet",
    concisename="We",
    buff=function(char) return char.combatname == "Poseidon" and 5 or 0 end
}

function M.wet.apply(target, combatstate)
    local aflame = statusops.statusduration(target, M.aflame)
    if aflame then
        statusops.removestatus(target, M.aflame, combatstate)
    end
    local twinserpents = statusops.statusduration(target, M.twinserpents)
    if twinserpents then
        statusops.removestatus(target, M.twinserpents, combatstate)
    end
    local warm = statusops.statusduration(target, M.warm)
    if wet then
        statusops.removestatus(target, M.warm, combatstate)
    end
end

--------------------- Warm -----------------------------------------
--[[
-- Warm doesn't do anything directly. Instead, Prometheus needs to be inflicted
-- with it for many of her powers to work, and her powers do
-- extra damage to anyone inflicted with it.
--
-- Though it does cancel hydroslide and wet.
--]]

M.warm = {
    name="warm",
    display="Warm",
    concisename="Wa",
    buff=function(char) return char.combatname == "Prometheus" and 5 or -1 end
}

function M.warm.apply(target, combatstate)
    local hydroslide = statusops.statusduration(target, M.hydroslide)
    if hydroslide then
        statusops.removestatus(target, M.hydroslide, combatstate)
    end
    local wet = statusops.statusduration(target, M.wet)
    if wet then
        statusops.removestatus(target, M.wet, combatstate)
    end
    local frozen = statusops.statusduration(target, M.frozen)
    if frozen then
        statusops.removestatus(target, M.frozen, combatstate)
    end
end

--------------------- Aflame -----------------------------------------
--[[
-- If Prometheus is inflicted with Aflame, her attack is doubled,
-- and she can use her most dangerous powers. The double damage is implemented
-- by the status, but unlocking powers will have to be done by her AI.
--
-- Anyone else inflicted with Aflame takes duration / 2 damage at the
-- end of their turn.
--]]

M.aflame = {
    name="aflame",
    display="Aflame",
    concisename="AF",
    notstackable=true,
    buff=function(char) 
        local manipulatesfire = char.combatname == "Prometheus" or char.combatname == "Flamey"
        return manipulatesfire and 12 or -5 
    end
}

function M.aflame.apply(combatstate, target, user)
    if target.name == user.name then
        local attackmodifiers = target.secondarystatistics.attackmodifiers
        attackmodifiers[M.aflame.name] = attacks.attackstat(target)
    end
end

function M.aflame.remove(target, combatstate)
    if M.aflame.buff(target) > 0 then
        local attackmodifiers = target.secondarystatistics.attackmodifiers
        attackmodifiers[M.aflame.name] = nil
    end
end

function M.aflame.endturn(target, combatstate)
    if M.aflame.buff(target) < 0 then
        local duration = statusops.statusduration(target, M.aflame).duration
        local damage = math.max(1, math.floor(duration / 2))
        damage = attacks.shield(target, damage)
        target.statistics.energy = target.statistics.energy - damage
        local combatname = target.combatname
        local description = F"Flames burn {combatname} for {damage} damage!"
        table.insert(combatstate.combatlog, description)
        if target.statistics.energy <= 0 then
            table.insert(combatstate.combatlog, F"{combatname} collapses!")
            combatstateops.defeatcharacter(combatstate, target.name)
        end
    end
end

-------------------------- Hydroslide --------------------------------------
--[[
-- The target receives a defense bonus equal to the status' duration, and an
-- extra Attack.
--]]
M.hydroslide = {
    name="hydroslide",
    display="Hydroslide",
    concisename="HS",
    buff=5
}

function M.hydroslide.apply(target, combatstate, duration)
    local choices = combatstate.choicecounts[target.name]
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    local buff = math.floor(attacks.statistic(target, "power") / 2)
    defensemodifiers[M.hydroslide.name] = buff
    choices.attack.current = choices.attack.current + 1
    choices.attack.max = choices.attack.max + 1
end

function M.hydroslide.remove(target, combatstate)
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    defensemodifiers[M.hydroslide.name] = nil
    local choices = combatstate.choicecounts[target.name]
    choices.attack.current = choices.attack.current - 1
    choices.attack.max = choices.attack.max - 1
end

-------------------- Twinserpents -------------------------------------------
--[[
-- Grants the target an additional power use.
--]]

M.twinserpents = {
    name="twinserpents",
    display="Twinserpents",
    concisename="TS",
    buff=10
}

function M.twinserpents.apply(target, combatstate)
    local choices = combatstate.choicecounts[target.name]
    choices.power.current = choices.power.current + 1
    choices.power.max = choices.power.max + 1
end

function M.twinserpents.remove(target, combatstate)
    local choices = combatstate.choicecounts[target.name]
    choices.power.current = choices.power.current - 1
    choices.power.max = choices.power.max - 1
end


--------------------- Tailwind -------------------------------------
--[[
-- Increases the target's damage by duration.
--]]

M.tailwind = {
    name="tailwind",
    display="Tailwind",
    concisename="TW",
    buff=1
}

function M.tailwind.apply(target, combatstate, user)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local buff = math.floor(attacks.statistic(user, "power") / 3)
    attackmodifiers[M.tailwind.name] = buff
end

function M.tailwind.remove(target, combatstate)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    attackmodifiers[M.tailwind.name] = nil
end

--------------------- Headwind -------------------------------------
--[[
-- Decreases the target's damage by the user's power / 3.
--]]

M.headwind = {
    name="headwind",
    display="Headwind",
    concisename="HW",
    buff=-1
}

function M.headwind.apply(target, combatstate, user)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local penalty = math.floor(attacks.statistic(user, "power") / 3)
    attackmodifiers[M.headwind.name] = -penalty
end

function M.headwind.remove(target)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    attackmodifiers[M.headwind.name] = nil
end

----------------- Stronger & Faster ----------
--[[
-- Provides the target a buff to attack and defense equal
-- to the user's power/10.
--]]
M.strongerfaster = {
    name="strongerfaster",
    display="Stronger & Faster",
    concisename="SF",
    buff=2
}
function M.strongerfaster.apply(target, combatstate, user)
    local buff = math.floor(attacks.statistic(user, "power") / 10)
    local modifiers = target.secondarystatistics
    modifiers.attackmodifiers[M.strongerfaster.name] = buff
    modifiers.defensemodifiers[M.strongerfaster.name] = buff
end

function M.strongerfaster.remove(target, combatstate)
    local modifiers = target.secondarystatistics
    modifiers.attackmodifiers[M.strongerfaster.name] = nil
    modifiers.defensemodifiers[M.strongerfaster.name] = nil
end

------------------ Bear Guard ---------------
--[[
-- Grants the target (Princess) a shield equal to the user's energy/2, and
-- increases the target's defense to match the user's. If the user is 
-- incapacitated, this status it prematurely ended at the start of the 
-- target's round. Reduces the user's energy by half.
--]]
M.bearguard = {
    name="bearguard",
    display="Bear Guard",
    concisename="BG",
    buff=1
}

function M.bearguard.apply(target, combatstate, user)
    local shieldbuff = math.floor(user.statistics.energy / 2)
    local userdefense = attacks.defensestat(user)
    local targetdefense = attacks.defensestat(target)
    local defensebuff = math.max(0, userdefense - targetdefense)
    local secondarystatistics = target.secondarystatistics
    secondarystatistics.defensemodifiers[M.bearguard.name] = defensebuff
    secondarystatistics.shield[M.bearguard.name] = shieldbuff
    user.statistics.energy = user.statistics.energy - shieldbuff
end

function M.bearguard.remove(target, combatstate)
    local shield = target.secondarystatistics.shield
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    defensemodifiers[M.bearguard.name] = nil
    shield[M.bearguard.name] = nil
end

function M.bearguard.startturn(target, combatstate)
    local activeallies = combatstateops.activeallies(combatstate)
    for _, character in pairs(activeallies) do
        if character.combatname == "Mr. Cuddles" then
            return
        end
    end
    statusops.removestatus(target, M.bearguard, combatstate)
end

------------------ Counter ---------------
--[[
-- The target receives 2 * attack in retaliation
-- points.
--]]
M.counter = {
    name="counter",
    display="Counter",
    concisename="C",
    buff=4
}

function M.counter.apply(target, combatstate)
    local buff = 2 * attacks.attackstat(target)
    target.secondarystatistics.retaliation[M.counter.name] = buff
end

function M.counter.remove(target, combatstate)
    target.secondarystatistics.retaliation[M.counter.name] = nil
end

------------------ Grabbed ---------------
--[[
-- The target has their attack and defense reduced by the user's strength/5 + 1,
-- and automatically recovers from Away
--]]
M.grabbed = {
    name="grabbed",
    display="Grabbed",
    concisename="G",
    buff=-7
}

function M.grabbed.apply(target, combatstate, user)
    local statusname = M.grabbed.name
    local targetname = assert(
        target.name,
        "Target does not have a name: " .. scarlettables.string(target)
    )
    local penalty = -(math.floor(attacks.statistic(user, "strength") / 5) + 1)
    target.secondarystatistics.attackmodifiers[M.grabbed.name] = penalty
    target.secondarystatistics.defensemodifiers[M.grabbed.name] = penalty
end

function M.grabbed.remove(character, combatstate)
    character.secondarystatistics.attackmodifiers[M.grabbed.name] = nil
    character.secondarystatistics.defensemodifiers[M.grabbed.name] = nil
end

function M.startturn(character, combatstate)
    if statusduration(character, M.away) then
        statusops.removestatus(character, M.away, combatstate)
        local msg = character.combatname .. " is yanked back into the fray!"
        table.insert(combatstate.combatlog, msg)
    end
end

--[[
-- A buff that modifies speed. It's called Fast if it
-- increases speed, and slow otherwise
--]]
function M.fast(amount)
    local fast = {
        name=amount > 0 and "fast" or "slow",
        display=amount > 0 and "Fast" or "Slow",
        concisename=amount > 0 and "Fa" or "Sl",
        buff=amount
    }
    function fast.apply(target, combatstate, duration, user)
        target.secondarystatistics.speedmodifiers[fast.name] = amount
    end

    function fast.remove(target, combatstate)
        target.secondarystatistics.speedmodifiers[fast.name] = nil
    end
    return fast
end

--[[
-- A buff that modifies strength. It's called Strong if it
-- increases strength, and weak otherwise
--]]
function M.strong(amount)
    local strong = {
        name=amount > 0 and "strong" or "weak",
        display=amount > 0 and "Strong" or "Weak",
        concisename=amount > 0 and "St" or "We",
        buff=amount
    }
    function strong.apply(target)
        target.secondarystatistics.strengthmodifiers[strong.name] = amount
    end
    function strong.remove(target)
        target.secondarystatistics.strengthmodifiers[strong.name] = nil
    end
    return strong
end

------------------ Reduce Attack ---------------
--[[
-- The target has their Attack reduced by 1. The
-- distinction with Away is that this doesn't trigger
-- any other statuses or modifiers that react to Away.
-- Another difference is that this status actually stacks.
-- Rather than increasing the duration of the status,
-- we instead further reduces the target's attack count.
-- Duration remains unaffected however.
--]]

local function reduceattack(target, combatstate)
    if not combatstate.statusstate[M.reduceattack.name] then
        combatstate.statusstate[M.reduceattack.name] = {}
    end
    local choices = combatstate.choicecounts[target.name]
    statusops.increasechoice(choices, "attack", -1)
    local statusstate = combatstate.statusstate[M.reduceattack.name]
    local reduction = statusstate[target.name] or 0
    statusstate[target.name] = reduction + 1
end

M.reduceattack = {
    name="reduceattack",
    display="Reduce Attack",
    concisename="RA",
    buff=-1,
    apply=reduceattack,
    stack=reduceattack
}

function M.reduceattack.remove(target, combatstate)
    local choicecountstarget = combatstate.choicecounts[target.name]
    local maxattackcount = choicecountstarget.attack.max
    local statusstate = combatstate.statusstate[M.reduceattack.name]
    local reduction = statusstate[target.name]
    local currentattackcount = choicecountstarget.attack.current
    choicecountstarget.attack.max = maxattackcount + reduction
    choicecountstarget.attack.current = currentattackcount + reduction
    statusstate[target.name] = nil
end

----------- Increase Spank
-- A status that increases the spank count by 1.
-- This stacks. So subsequent applications will further increase
-- the spank count.

function M.increasespank(possibledescriptions)
    local increasespanking = {
        name="increasespank",
        display="Increase Spank",
        concisename="S",
        buff=1,
    }

    function increasespanking.apply(target, combatstate)
        if not combatstate.statusstate[increasespanking.name] then
            combatstate.statusstate[increasespanking.name] = {}
        end
        local statusstate = combatstate.statusstate[increasespanking.name]
        local currentreduction = statusstate[target.name] or 0
        statusstate[target.name] = currentreduction + 1
        local choices = combatstate.choicecounts[target.name]
        statusops.increasechoice(
            choices,
            "spank",
            1,
            scarletchoices.spank(possibledescriptions)
        )
    end

    increasespanking.stack = increasespanking.apply

    function increasespanking.remove(target, combatstate)
        local statusstate = combatstate.statusstate[increasespanking.name]
        local reduction = statusstate[target.name]
        statusstate[target.name] = nil
        local choices = combatstate.choicecounts[target.name]
        statusops.decreasechoice(choices, "spank", reduction)
    end

    return increasespanking
end

----------- Twister
--[[
-- Any character but Tempestas inflicted with Twister takes damage at the
-- start of their turn equal to away.duration * tempestas.power/2.
-- At the start of her turn, Tempestas is cured of Away and receives a
-- one turn attack bonus Away.duration * tempestas.power/2.
--]]

M.twister = {
    name="twister",
    display="Twister",
    concisename="T",
    buff=function(character) 
        return string.find(character.name, "tempestas") and 40 or -1
    end
}

function M.twister.apply(character, combatstate, user)
    if not combatstate.statusstate then
        combatstate.statusstate = {}
    end
    if not combatstate.statusstate[M.twister.name] then
        combatstate.statusstate[M.twister.name] = {}
    end
    combatstate.statusstate[M.twister.name][character.name] = user
end

function M.twister.remove(character, combatstate)
    combatstate.statusstate[M.twister.name][character.name] = nil
end

function M.twister.startturn(character, combatstate)
    local away = statusops.statusduration(character, M.away)
    local user = combatstate.statusstate[M.twister.name][character.name]
    local activecharacter = combatstateops.activecharacter(combatstate)
    if string.find(character.name, "tempestas") and away then
        if activecharacter.name == character.name then
            local buff = math.floor(attacks.statistic(character, "power") / 2)
            local awayduration = away and away.duration or 0
            local attackmods = character.secondarystatistics.attackmodifiers
            attackmods[M.twister.name] = awayduration * buff
            table.insert(
                combatstate.combatlog,
                character.combatname .. " slingshots back into the fray!"
            )
            statusops.removestatus(character, away.status, combatstate)
        end
    else
        if away then 
            local damage = math.floor(attacks.statistic(user, "power") / 2)
            if statusops.statusduration(character, M.fly) then
                damage = damage * 2
            end
            damage = attacks.shield(character, away.duration * damage)
            character.statistics.energy = character.statistics.energy - damage
            if damage > 0 then
                table.insert(
                    combatstate.combatlog,
                    F"{character.combatname} is whipped around by the twister, "
                        .. F"taking {damage} damage!"
                )
            end
            statusops.removestatus(character, away.status, combatstate)
            if character.statistics.energy <= 0 then
                combatstateops.defeatcharacter(combatstate, character.name)
                table.insert(
                    combatstate.combatlog,
                    F"{character.combatname} collapses!"
                )
            end
        end
    end
end

function M.twister.endturn(character, combatstate)
    if string.find(character.name, "tempestas") then
        local attackmodifiers = character.secondarystatistics.attackmodifiers
        attackmodifiers[M.twister.name] = nil
    end
end

----------- S-Armor -----
--[[
-- A character who has been inflicted with S-Armor has their
-- shield statistic incremented by 2 * user.statistics.power.
-- If at the start of a character's turn their shield is at 0,
-- the shield status is removed. It also increases attrition
-- by 1.
--]]
M["s-armor"] = {
    name="s-armor",
    display="S-Armor",
    concisename="SA",
    buff=1,
    remove=function() end
}

M["s-armor"].apply = function(target, combatstate, user)
    local buff = 2 * attacks.statistic(user, "power")
    target.secondarystatistics.shield[M["s-armor"].name] = buff
    statusops.increaseattrition(combatstate, target, 1)
end

M["s-armor"].remove = function(target, combatstate)
    target.secondarystatistics.shield[M["s-armor"].name] = nil
    statusops.increaseattrition(combatstate, target, -1)
end

M["s-armor"].stack = function(target, combatstate, user)
    local buff = 2 * attacks.statistic(user, "power")
    local damage = target.secondarystatistics.shield.damage
    target.secondarystatistics.shield.damage = math.min(0, damage + buff)
    target.secondarystatistics.shield[M["s-armor"].name] = buff
end

M["s-armor"].startturn = function(character, combatstate)
    if attacks.shield(character) <= 0 then
        table.insert(
            combatstate.combatlog, 
            F"{character.combatname}'s shield crumbles!"
        )
        statusops.removestatus(character, M["s-armor"], combatstate)
    end
end

----------- Wall
--[[
-- A character who has been inflicted with Wall has all damage taken and
-- done by them reduced for 1 turn.
--]]
M.wall = {
    name="wall",
    display="Wall",
    concisename="W",
    buff=1
}

function M.wall.apply(target, combatstate, user, buff)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    attackmodifiers[M.wall.name] = -buff
    defensemodifiers[M.wall.name] = buff
end

function M.wall.remove(target)
    local attackmodifiers = target.secondarystatistics.attackmodifiers
    local defensemodifiers = target.secondarystatistics.defensemodifiers
    attackmodifiers[M.wall.name] = nil
    defensemodifiers[M.wall.name] = nil
end

--------- Reared
--[[
-- A character who has been inflicted with reared has their speed doubled
-- for one round.
--]]
M.reared = {
    name="reared",
    display="Reared",
    concisename="R",
    buff=1
}

function M.reared.apply(character, combatstate)
    local speed = attacks.statistic(character, "speed")
    local modifiers = character.secondarystatistics.speedmodifiers
    modifiers[M.reared.name] = speed
end

function M.reared.remove(character, combatstate)
    local modifiers = character.secondarystatistics.speedmodifiers
    modifiers[M.reared.name] = nil
end

--------- Speed
--[[
-- Speed modifies a character's speed. Unlike the other statuses,
-- this is a actually a function that takes the amount to modify
-- speed by and returns the status to apply.
--]]

function M.speed(modifier)
    local stat = {
        name=modifier > 0 and "haste" or "slow",
        display=modifier > 0 and "Haste" or "Slow",
        concisename=modifier > 0 and "H" or "S",
        buff=modifier
    }

    function stat.apply(character, combatstate)
        local modifiers = character.secondarystatistics.speedmodifiers
        modifiers[stat.name] = modifier
    end

    function stat.remove(character, combatstate)
        local modifiers = character.secondarystatistics.speedmodifiers
        modifiers[stat.name] = nil
    end
end


------- Lego Swirl
--[[
-- Reduces the target's attack by user.statistics.strength / 5.
--]]
M.legoswirl = {
    name="legoswirl",
    display="Legoswirl",
    concisenme="LS",
    buff=-5
}

function M.legoswirl.apply(character, combatstate, user)
    local reduction = math.floor(attacks.statistic(user, "strength") / 5)
    local attackmodifiers = character.secondarystatistics.attackmodifiers
    attackmodifiers[M.legoswirl.name] = -reduction
end


function M.legoswirl.remove(character, combatstate)
    local attackmodifiers = character.secondarystatistics.attackmodifiers
    attackmodifiers[M.legoswirl.name] = nil
end

------ Stasis
--[[
-- Cures the target of the worst status (except this one) currently inflicting that
-- character, but does damage equal to the duration of the cured status.
--]]

M.stasis = {
    name="stasis",
    display="Stasis",
    concisenme="St",
    buff=20
}

function M.stasis.apply(character, combatstate, user)
    -- Stasis does nothing when it's first applied.
end


function M.stasis.remove(character, combatstate, user)
    -- Stasis does nothing when it's removed.
end

function M.stasis.startturn(character, combatstate)
    local worstbuffvalue = 10000
    local worststatus
    local energy = character.statistics.energy
    for i, statusduration in ipairs(character.statuses) do
        local status = statusduration.status
        local duration = statusduration.duration
        if status.name ~= M.stasis.name and duration < energy then
            local buff = status.buff
            if type(buff) == "function" then
                buff = buff(character)
            end
            if buff < worstbuffvalue then
                worstbuffvalue = buff
                worststatus = statusduration
            end
        end
    end
    if worststatus then
        local status = worststatus.status
        local duration = worststatus.duration
        table.insert(
            combatstate.combatlog,
            F"{status.name} is purged from {character.combatname}! "
                .. F"{character.combatname} takes {duration} damage!"
        )
        statusops.removestatus(character, status, combatstate)
        character.statistics.energy = energy - duration
    end
end

--[[
-- Transfers all of a character's stat points into a particular stat.
-- At the start of a character's turn, the points get moved to a different
-- stat
--]]

M.distortion = {
    name="distortion",
    display="Distortion",
    concisenme="Di",
    buff=-50
}

local function cleardistortion(character) 
    local modifiers = character.secondarystatistics
    modifiers.strengthmodifiers[M.distortion.name] = nil
    modifiers.speedmodifiers[M.distortion.name] = nil
    modifiers.powermodifiers[M.distortion.name] = nil
end

local function distort(character, stat)
    cleardistortion(character)
    local modifiers = character.secondarystatistics
    local strength = attacks.statistic(character, "strength")
    local speed = attacks.statistic(character, "speed")
    local power = attacks.statistic(character, "power")
    modifiers.strengthmodifiers[M.distortion.name] = -strength
    modifiers.speedmodifiers[M.distortion.name] = -speed
    modifiers.powermodifiers[M.distortion.name] = -power
    if stat == "strength" then
        modifiers.strengthmodifiers[M.distortion.name] = speed + power
    elseif stat == "Speed" then
        modifiers.speedmodifiers[M.distortion.name] = strength + power
    else
        modifiers.powermodifiers[M.distortion.name] = strength + speed
    end
end

function M.distortion.apply(character, combatstate, user)
    -- Nothing happens when distortion is applied. It triggers at
    -- the start of each turn.
end

function M.distortion.startturn(character, combatstate)
    if attacks.statistic(character, "strength") > 0 then
        distort(character, "speed")
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} feels incredibly fast!"
        )
    elseif attacks.statistic(character, "speed") > 0 then
        distort(character, "power")
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} feels incredibly fast!"
        )
    else
        distort(character, "strength")
        table.insert(
            combatstate.combatlog,
            F"{character.combatname} feels incredibly strong!"
        )
    end
end

function M.distortion.remove(character, combatstate)
    cleardistortion(character)
    table.insert(combatstate.combatlog, F"{character.combatname} stabilizes!")
end

------------------------ Attack ----------------------------
-- [[
-- The Attack status increases the number of times the character can attack by 1.
-- ]]
M.attack = {
    name="Attack",
    display="attack",
    concisename="AT",
    buff=2
}

function M.attack.apply(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    statusops.increasechoice(choices, "attack", 1, scarletchoices.attack)
    statusops.increaseattrition(combatstate, character, 1)
end

function M.attack.remove(character, combatstate)
    local choices = combatstate.choicecounts[character.name]
    if choices.attack then
        removechoice(choices, "attack")
    end
    statusops.increaseattrition(combatstate, character, -1)
end


--------------------- Fly ---------------------------------
M.fly = {
    name="Fly",
    display="fly",
    concisename="F",
    buff=10
}

function M.fly.apply(character, combatstate, user, buff) 
    character.secondarystatistics.attackmodifiers[M.fly.name] = buff
    character.secondarystatistics.defensemodifiers[M.fly.name] = buff
    statusops.increaseattrition(combatstate, character, 1)
end

function M.fly.remove(character, combatstate, user, buff)
    character.secondarystatistics.attackmodifiers[M.fly.name] = nil
    character.secondarystatistics.defensemodifiers[M.fly.name] = nil
    statusops.increaseattrition(combatstate, character, -1)
end


--------------------- Vulnerable ---------------------------------
--[[
-- Vulnerable reduces the target's defense by the specified amount.
--]]
M.vulnerable = {
    name="Vulnerable",
    display="vuln",
    concisename="V",
    buff=-5
}

function M.vulnerable.apply(character, combatstate, user, buff) 
    character.secondarystatistics.defensemodifiers[M.vulnerable.name] = -buff
end

function M.vulnerable.remove(character, combatstate, user, buff)
    character.secondarystatistics.defensemodifiers[M.vulnerable.name] = nil
end

return M
