--[[
    Copyright (C) 2019  AKA Russell

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[
--This is a simple parser that takes in a Scarlet story file, which use a
--subset of org-mode syntax, and returns a table that can be understood by
--the rest of the Scarlet game engine. A Scarlet story file should be
--structured as follows:
--
--* event_name
--Events are delimited using the header-1 syntax (i.e. a single '*'). Text that
--should appear in a story appears immediately below it. Subsequent options, like
--`do`, `automatics`, and `choices` appear as sub-headings ('**') in any order.
--
--The body of a story is stored as a single string with all formatting preserved, which
--Scarlet splits apart and reformats as needed.
--
--Choices are choices the player can make. They appear as org-style links to other
--events (so the game could even be played as a CYOA using html). automatics are
--similar, except they are choices made by the engine, based on the current state 
--of the game. The `do` heading is a list of commands the engine should execute,
--which will modify the game state.
--
--
--** choices
--- [[#event_name_2][I'd like to go to event_name_2]]
--- [[#event_name_3][I'm gonna go to event_name_3]]
--
--** automatics
--- [[#event_name_2][predicate;argument]]
--- [[#event_name_3][next_predicate;argument]]
--- [[#event_name_4][true]]
--
--
--** do
-- - operation;argument
--
-- * event_name_2
-- ** allies
-- - pc
-- ** enemies
-- - thug
-- - buzzsaw
-- ** victory
-- event_name_3
-- ** defeat
-- event_name_4
--
--This will get translated into the following Lua table:
--{
--    event_name_1={
--        story= "Events are delimited using the header-1 syntax (i.e. a single '*'). Text that..."
--        choices= {
--            "I'd like to go to event_name_2"="event_name_2",
--            "I'm gonna go to event_name_3"="event_name_3",
--            _keys={"I'd like to go to event_name_2", "I'm gonna go to event_name_3"}
--        },
--        automatics={
--            "predicate;argument": "event_name_2",
--            "next_predicate;argument": "event_name_3",
--            "true": "event_name_4"
--        },
--        do={"operation;argument"}
--    },
--    event_name_2={
--        allies={"pc"},
--        enemies={"thug", "buzzsaw"},
--        victory="event_name_3",
--        defeat="event_name_4"
--    }
--}
----]]

local M = {}

local function tokenize(line)
    local tokens = {}
    for token in string.gmatch(line, "[^%s]+") do
        table.insert(tokens, token)
    end
    return tokens
end

local function slice(list, first, last)
    first = first or 1
    last = last or #list
    local sliced = {}
    for i=first,last do
        table.insert(sliced, list[i])
    end
    return sliced
end

local function invalid_event_name(event_name, linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "Invalid event name '"
        .. event_name
        .. "' is reserved for event subheadings. Did you forget a '*'?"
end

local function invalid_sub_heading(heading, linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "Unrecognized subheading '"
        .. heading
        .. "'."
end

local function invalid_link(link, linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "Malformed link: '"
        .. link
        .. "'. Expected syntax is [[event_name][expression]]."
end

local function undefined_event(line, linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "No event defined before '" .. line
        .. "'Did you forget a single '*' before the event name?"
end

local function bare_line(linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "Bare line (i.e. a line without a leading symbol) "
        .. "encountered outside the story body, a victory or defeat subheading."
end

local function missing_space(line, linenum, filename)
    return filename .. ":" .. linenum .. ": "
        .. "Malformed heading: '" .. line
        .. "'. Make sure there's a space between the asterix and your event name."
end

local subheadings = {
    choices=1,
    automatics=1,
    ["do"]=1,
    allies=1,
    enemies=1,
    victory=1,
    defeat=1,
    leaf=1
}

--[[
Given an iterator of org-mode lines, returns a lua table that mimics the
structure. This isn't a proper parser of org-mode though. It only supports
the features I need for my game, and it does some sanity checks. For example,
top-level events can't be called "do" "automatics", "choices" "allies" or
"enemies" in any sort of casing because those are reserved for event
sub-headings.

orglines: The lines of text containing an org-mode definition of the game
filename: The name of the file being processed, used for useful error messages
--]]
function M.decode(orglines, filename)
    local linenum = 1
    local events = {_keys={}}
    local errors = {}
    local event
    local currenttable
    local heading
    for line in orglines do
        local tokens = tokenize(line)
        local non_empty = #tokens > 0
        if non_empty and tokens[1].sub(1, 1) == "#" then
            -- We skip any lines that start with #, because those are comments.
        elseif non_empty and tokens[1] == "*" then
            local event_name = table.concat(slice(tokens, 2), " ")
            event = {story={}}
            currenttable = event.story
            heading = "story"
            events[event_name] = event
            table.insert(events._keys, event_name)
            if subheadings[event_name] then
                table.insert(errors, invalid_event_name(event_name, linenum, filename))
            end
        elseif non_empty and tokens[1] == "**" then
            if not event then
                table.insert(errors, undefined_event(heading, linenum, filename))
            end
            if currenttable and currenttable == event.story then
                event.story = next(event.story) and table.concat(event.story, "\n") or nil
            end
            heading = tokens[2]
            if not subheadings[heading] then
                table.insert(errors, invalid_sub_heading(heading, linenum, filename))
            end
            currenttable = heading ~= "victory" and heading ~= "defeat" and heading ~= "leaf"
                and {}
                or nil
            event[heading] = currenttable
        elseif non_empty and tokens[1] == "-" then
            if heading == "victory" or heading == "defeat" then
                table.insert(
                    errors, 
                    "Victory and Defeat headings don't take lists. Bad event: " .. event_name
                )
                break
            end
            local content = table.concat(slice(tokens, 2), " ")
            if string.sub(tokens[2], 1, 1) == "[" then
                local eventnameexpression = string.gmatch(content, "%[([^%[%]]+)%]")
                local eventname = eventnameexpression()
                assert(eventname, "Eventname not found for " .. line)
                if string.sub(eventname, 1, 1) == "#" then
                    eventname = string.sub(eventname, 2)
                end
                local expression = eventnameexpression()
                if not eventname or not expression then
                    table.insert(errors, invalid_link(content, linenum, filename))
                else
                    currenttable[expression] = eventname
                    if not currenttable._keys then
                        currenttable._keys = {}
                    end
                    table.insert(currenttable._keys, expression)
                end
            else
                table.insert(currenttable, content)
            end
        else
            if non_empty and tokens[1].sub(1, 1) == "*" then
                table.insert(errors, missing_space(line, linenum, filename))
            end
            if currenttable then
                if currenttable ~= event.story and non_empty then
                    table.insert(errors, bare_line(linenum, filename))
                elseif currenttable == event.story then
                    table.insert(currenttable, line)
                end
            elseif non_empty then
                assert(event, "No event associated with " .. line)
                event[heading] = line
            end
        end
        linenum = linenum + 1
    end
    assert(
        #errors == 0,
        "Encountered errors parsing story: \n" .. table.concat(errors, "\n")
    )
    return events
end

return M
