--[[
-- Copyright 2016 AKA Russell
--
-- This file is part of Scarlet.
--
-- Scarlet is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- Scarlet is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with Scarlet.  If not, see <http://www.gnu.org/licenses/>.
--]]

--[[ An page is a table consisting of an page body, the next M to
-- transition to, and a function that transforms the gamestate.
--
-- There are two flavors of next M: automatic and
-- choice. The Automatic is a list of pairs. The first entry of each pair
-- is a predicate over the game state, while the second is the M associated
-- with each predicate. We automatically
-- transition to the first automatic page to evaluate to true.
-- If they all evaluate to false (or there aren't any), then we instead
-- present the player with the choices from the choices table.
--
-- The choices table is a list of pairs. The first entry in each pair is the
-- String to display to the user, the second is the page to transition to if
-- the player makes that choice.
--]]

local M = {}
local log = require 'log'

M.AUTOMATIC_EVENT = 1
M.CHOICE_TEXT = 1
M.EVENT = 2

local function noautomatics()
    return nil
end

local function nochoices()
    return {_keys={}}
end

local function notransformer() end

local function nobody()
    return {}
end

--[[
-- Given the body of an page, a function that maps state to event
-- (representing events that should be transitioned to automatically), and a
-- function that takes a state and returns a table of player choices, returns
-- a page.
--
-- The body is a function that takes a state, and returns a list of
-- Strings that will be displayed to the player when the page is played.
-- Defaults to a function that returns an empty table
--
-- The automatics is a function that takes a state and returns another page.
--
-- The choices is a function that takes a state and returns an ordered table
-- from string to page, representing the choices the player may make.
--
-- The transformer is an action that takes the current state of the
-- game, modifies it as appropriate for this page. If nil, defaults to the
-- no-op action.
--
-- Every parameter except for body is optional.
-- ]]
function M.page(body, automatics, choices, transformer)
    return {
        body = body or nobody,
        automatics = automatics or noautomatics,
        choices = choices or nochoices,
        transformer = transformer or notransformer
    }
end

--[[
-- Given an page, the current state of the game, and an index representing the
-- player's choice, returns the next page to display
--]]
function M.nextChoice(page, state, choice)
    local choices = page.choices(state)
    return choices[choices._keys[choice]]
end

return M
