# Table of Contents #

- [Overview](#frontend)

- [Writing Pages](#writing-pages)
    * [Templated Text](#templated-text)
    * [Predicates](#predicates)
    * [Operations](#operations)

- [UI](#ui)

- [Keywords](#keywords)

- [Lua Structure](#lua-structure)

- [Combat](#combat)

- [Loading](#loading)
    * [Resolving Templated Text](#resolving-dynamic-text)

# Overview #

Here, we define story pages in detail, including:

1. How to define them in Lua.

2. Their appearance in the UI

3. Their internal structure in Lua.

4. Our strategy for loading them from YAML into Lua tables.


Story pages encode the story itself: they include such things as the text to
display, any music that should be played while a given page is displayed, the
choices that transition to additional pages, and how to track events through the
story.

This does NOT deal with how to display story contents. That is defined in
[UI Story Pages](./pages.md#story-ui-page).

# Writing Pages #

Story pages are stored in `data/stories`, and defined in YAML.

Story pages may be stored in any number of YAML files.

The pages themselves are defined with the following structure:
```yaml
hunger:
  story: >
    [name;greg] looked out across the land and declared: "I am hungry."
  automatics:
    hasKeyword(gregory, gluttonous): start_eating
    hasKeyword(gregory, spiritual): start_meditating
  choices:
    Go down to eat something.: start_eating
    Meditate.: start_meditating
  do:
    - modifyStatistic;gregory; hunger; 10
```

`story`, and `choices` are [templated text](#templated-text).

`story` defines the body of each event, which will be displayed to the player.

`choices` defines conscious decisions the player can make to decide how to
advance the story. `choices` will be numbered and displayed to the player in the
order defined in the page.

`automatics` describes under what conditions (if any) the game should
automatically transition to a different page. The conditions are defined using
[predicates](#predicates).

Note: Cycles are not allowed in the automatic choices, meaning that you cannot
automatically transition from page `x` to page `y` and then back to page `x`.
This will be verified at load time.

`do` describes the [operations](#operations) that should be applied to the
gamestate when leaving the page. For example, the writer may choose to
modify character statistics or add a keyword. The operations are applied in the
order they are defined in the `change` block.


## Templated Text ##

In addition to raw text, story writers also have access to a collection of
functions that allow the writer to display different text depending on the
state of the game, or a character. This is called templated text. Both the
story body and the choices text can be templated.

Functions are wrapped in square brackets
in the text. For example, suppose we would like to allow the player to change
Gregory's name in the event above. Then our event would be:

```yaml
hunger:
  music: dreadful_music
  story: >
    [name;greg] looked out across the land and declared: "I am hungry."
  automatics:
    hasKeyword(gregory, gluttonous): start_eating
    hasKeyword(gregory, spiritual): start_meditating
  choices:
    Go down to eat something.: start_eating
    Meditate.: start_meditating
  do:
    - modifyStatistic;gregory; hunger; 10

start_eating: ...
start_meditating: ...
```

Observe that Gregory's name has been replaced by `[name;greg`]. `name` is a
function that takes the id of a character (`greg` in this case) and returns
the name for the character that will be displayed to the player. So suppose
the player named `greg` `Jones`. Then, we would display the following ingame:

    `Jones looked out across the land and declared: "I am hungry."`

Template functions can take any number of required and optional parameters (how
many, and which are optional and which are required depend on the function).
Parameters are separated by the semi-colon (`;`). The first parameter is the
name of the function to invoke

There is a whole mess of template functions available in the
[template_functions](./template_functions.md) file, and all template function
are defined in `story/templatefunctions.lua`. However, they are all merely
syntactic sugar around one or more of the following functions:

### attribute ###

This function displays text depending on whether the specified character has
the specified value for the specified attribute. For example,
`attribute;greg; gender; male; he; she` will return `he` if the `greg`'s
gender is `male`, and will return `she` otherwise.

If only the character and attribute name are specified, then this function
will return the value the specified character has for the specified attribute.
For example, suppose `greg` is `male`. Then, `attribute(greg; gender)` will
return `male`.

Signature:
    `attribute;character; attributename; value; textifequal; textotherwise`
    `attribute;character; attributename`
Parameters:
    - `character` - The character whose attribute should be checked
    - `attributename` - The name of the attribute to check
    - `value` - The value to compare the character's attribute against
    - `textifequal` - The text to print of the character's attribute is
        equal to `value`.
    - `textotherwise` - The text to print otherwise


### keyword ###

This function will print text depending on whether or not the player has
acquired the specified [keyword](#keywords). For example, the function
`keyword;the_thing; I've got the thing!; I don't have the thing...`
will return `I've got the thing!'` if the player has the keyword
`the_thing` and `I don't have the thing` if the player does not have
the keyword `the_thing`.

Signature:
    `keyword;keywordname; textifhaskeyword; textotherwise`

Parameters:
    - `keywordname` - The keyword the player may or may not have
    - `textifhaskeyword` - The text to return if the player has the specified
        keyword
    - `textotherwise` - The ext to return if the player does not have the
        specified keyword

### statistic ###

This function will print text depending on the relative value of a statistic.
For example, `statistic;greg; hunger; gt; 5; I am hungry!; I am not...` will
return `I am hungry!` if `greg`'s `hunger` is greater than `5`, and
`I am not...` otherwise.

If only the `character` and `statisticname` are specified, then this function
returns the current value of the specified statistic. So, suppose `greg` has
a `hunger` of `5`. Then, `statistic(greg; hunger)` returns `5`.

Signature:
    `statistic;character; statisticname; comparison; value; textiftrue; textiffalse`
    `statistic;character; statisticname`

Parameters:
    - `character` - The name of the character whose statistic should be checked
    - `statisticname` - The statistic of interest
    - `comparison` - The comparison to perform. This can have one of the
        following values:
        * `gt` - Greater than
        * `lt` - Less than
        * `e` - Equal
        * `gte` - Greater than or equal
        * `lte` - Less than or equal
    - `value` - The value to compare the character's statistic against.
    - `textiftrue` - The text to print of the specified character's statistic
        compares favorably against the specified value.
    - `textiffalse` - The text to print otherwise.

## Predicates ##

Predicates are functions that can be used when defining `automatics`. There
is a whole mess of them available [predicates](./predicates.md) document.
However, like with templated text functions, they are all syntactic sugar
for combinations of the following atomic predicates:

### attribute ###

`attribute` evaluates to true if the specified character has the specified
value for the specified attribute. For example,
`attribute(greg; gender; male)` will return `true` if the character `greg` is
`male`, and false otherwise.

Signature:
    `attribute;character; attributename; value`

Parameters:
    - `character` - The name of the character whose attribute needs to be
        checked
    -  `attributename` - The name of the attribute that should be checked.
    -  `value` - The value that the character's attribute should be compared
        against

### keyword ###

`keyword` evaluates to true if the player has the specified keyword. For
example, `keyword(gluttonous)` will return true if the player has the
`gluttonous` keyword, and false otherwise.

Signature:
    `keyword(keywordname)`

Parameters:
    - `keywordname` - The name of the keyword the player should possess.

### statistic ###

Checks the relative value of the specified statistic for the specified
character. For example, `statistic;greg; hunger; gt; 10` returns `true` if
`greg` has more than `10` `hunger`.

Signature:
    `statistic;character; statisticname; comparison; value`

Parameters:
    - `character` - The name of the character to be tested
    - `statisticname` - The name of the statistic to compare
    - `comparison` - The comparison to perform. This can have one of the
        following values:
        * `gt` - Greater than
        * `lt` - Less than
        * `e` - Equal
        * `gte` - Greater than or equal
        * `lte` - Less than or equal
    - `value` - The value to compare the character's statistic against

## Operations ##

Operations are transformations applied to the game's state when leaving a page.
There is a whole mess of operations in [operations](./operations.md), but
they are all syntactic sugar around one or more invocations of the following
atomic operations.

Operations are defined in `story/operations.lua`.


### attribute ###
`attribute` changes the specified attribute on the specified character to the
specified value. For example `attribute;greg; gender; female` will make `greg`
a woman.

Signature:
    `attribute;character; attributename; value`
Parameters:
    - `character` - The name of the character whose attribute should be changed
    - `attributename` - The name of the attribute to be changed
    - `value` - The value to change the attribute to

### keyword ###
`keyword` adds the specified keyword to the player's set of acquired keywords.

Signature:
    `keyword;keywordname`

Parameters:
    - `keywordname` - The keyword to add

### statistic ###
`statistic` sets the specified statistic on the specified character to the
specified value. For example `statistic;greg;hunger;10` sets `greg`'s
hunger statistic to 10.

Signature:
    `statistic;character; statisticname; value`

Parameters:
    - `character` - The character whose statistic should be modified
    - `statisticname` - The name of the statistic to be modified
    - `value` - The value to set the specified statistic to

# UI #

ASCII sketch of the UI:

------------------------------------------------------------
| Text is large, and each line is sixty characters long.   |
| More text, because text is like awesome and junk.        |
| 1. Make a choice.                                        |
| 2. Another choice                                        |
| 3. So many choices!                                      |
------------------------------------------------------------
| (#) Response (↑/↓) scroll (PgUp/PgDown) Page Up/Down
| (S)ave/Load (Q)uicksave (C)haracter (O)ptions (T)itle Screen |

Note: The right hand of the screen will pretty much be wasted
with only sixty characters per line. So we may want to add a panel to the right
that displays the player's current status:

------------------------------------------------------------
| Text is large, and each line is sixty characters long.   |  Character:
| More text, because text is like awesome and junk.        |  Energy: 45/45
|                                                          |  Grapple: 10
| 1. Make a choice.                                        |  Armslength: 20
| 2. Another choice                                        |  Power: 15
| 3. So many choices!                                      |  Armor: T-Shirt
|                                                          |  Weapon: Fists
------------------------------------------------------------
(#) Response (*up*/*down*/) scroll lines (PgUp/PgDown) Page Up/Down
(S)ave/Load (Q)uicksave (C)haracter (O)ptions (T)itle Screen


However, this will almost certainly require drawing a line on the right margin
to provide some visual separation.

# Keywords #

Keywords are arbitrary strings a developer can use to track progress through
the story. For example, suppose the player character (PC) coming to a fork
in the road, one going north and one going south, each with a
different story. Eventually, you may want the two paths to converge again,
but still do different things later depending on which path they took (maybe
they meet and befriend strange witch doctor on the south path, and a grumpy
ogre on the north). You could track this with by adding a keyword `southpath`
if they took the south path, and then check for that keyword whenever you
want to do something different.

# Lua Structure #

Like all pages, story pages are Lua tables with the following mappings:

- `body`: A function that takes the current gamestate, and returns a list of
    paragraphs to display. These paragraphs use the `gamestate` to resolve
    the templated text defined in the [yaml](writing-pages) into
    text to display to the player.
- `choices`: A function that takes the current gamestate, and returns an
    ordered table that maps text (representing player choices) to pages.
    If desired, a writer may wrap a single character in parenthesis. This
    character will be used to select that choice, rather than a number
    (provided by default).
- `automatics`: A function that takes the current gamestate, and returns
    the next page to display, or `nil` if there isn't one.
- `transformer`: An operation that transforms the current gamestate.

They also have the additional mapping:

- `name`: The name of this story event


Note: The `transformer` is NOT the same as the `do` block defined above. The
`transformer` is applied when transitioning into a page, NOT when leaving it.
As a result the `transformer` will in fact be executing operations from the
previous page. So suppose we have page `A` that can transition to page `B`, and
suppose page `A` has operation `X` defined. Then, `X` will be embedded in
`B`'s `transformer`. This embedding will occur when building the table
returned by `choices`, or the page returned by `automatics`.


# Loading #

When the game loads, story pages are loaded into the `configuration.pages`
table, a table of story pages keyed on name. However, these pages are not what
are actually returned by other `choices` and `automatics` functions. Suppose
we have three pages `a`, `b`, and `c` defined as follows:

```yaml
a:
    text: >
        aaaaa
    choices:
        bbb: b
        ccc: c
    do:
        - doA()
b:
    text: >
        bbbbb
c:
    text: >
        cccc
```

Recall that `doA` needs to be performed when leaving `a`, but the `transform`
function is applied when entering a page. In other words, we need to invoke
`doA` when entering `b` if we choose `bbb` and when entering `c` when invoking
`ccc`.

Therefore, `a`'s choice function will build a new page from the templates for
`b` and `c` that includes a `transform` function that applies `doA`. Something
like:

```lua
    function a_choices(gamestate)
        local btemplate = gamestate.configuration.stories.b
        local ctemplate = gamestate.configurations.stories.c
        local transform = gamestate.configuration.operations.doA
        local bpage = pages.page(
            btemplate.body,
            btemplate.automatics,
            btemplate.choices,
            transform
        )
        local cpage = pages.page(
            ctemplate.body,
            ctemplate,automatics,
            ctemplate.choices,
            transform
        )
        return {b=bpage, c=cpage, keys={b,c}}
    end
```

In other words, we essentially inject the `do` block of `a` into the
`transform` function of the story pages of `b` and `c`.

The exact same thing happens with automatics.


## Resolving Templated Text ##

A key part of displaying text to the player based on game state is taking the
templated text defined in the story files and resolving them into concrete
text. We do this in two steps:

1. Translate the template text into a data structure that separates raw
    text that should be displayed as-is, and the templated text.
2. Replace the templated text with a function that takes the current gamestate
    and returns a string to display to the user.

Obviously, we may have a lot of text (thousands of lines) and we need to be
very fast about reducing templated text into concrete text. Furthermore,
most of the text will be static, rather than templated. Finally, the templated
text may of course be recursive (template functions may take templated text
to display).

Therefore, we split templated text into a table of tables. Each table represents
a paragraph. Each entry in each paragraph is either raw
text, or a function on gamestate. Then, when we need to make the text
concrete, we create a copy of the table of paragraphs. For each entry, in
each paragraph, if the entry is
a string, then it is copied into the new table. If it is a function, it is
evaluated and the result added to the new table. Then we concatenate
paragraphs together and return a list of paragraphs describing the concrete
text.

For example, consider the following text:

```
Testing lots of [hisher] paragraphs.

I am a fool! [HeShe;robert] declared.
```

This template would be parsed into the following table:

```lua
{
    {"Testing lots of ", templatefunctions.hisher, " paragraphs."},
    {
        "I am a fool! ",
        function(gamestate)
            return templatefunctions.HeShe(gamestate, "robert")
        end,
        " declared."
    }
}
```

# Combat #

Combat is unfortunately going to require its own unique structure, because
this is less about displaying something to the user, and more about telling
the system what it needs for combat.

Combat's structure is the following:

```yaml

name:
    allies:
        - ally1
        - ally2
    enemies:
        - enemy1
        - enemy2
    victory: victorypagename
    loss: losspagename
```

We need to define three things:
    1. The characters involved
    2. Their starting positions
    3. The post combat pages.
